"""Preconfigured converters for msgpack."""

from cattrs.preconf.msgpack import MsgpackConverter, configure_converter, make_converter

__all__ = ["MsgpackConverter", "configure_converter", "make_converter"]
